@@ -15,11 +15,11 @@ Basic use:
 
 If you know the delimiters, qualifiers, etc, you may skip the optional
 'guessing' steps as they rely on heuristics anyway (although they seem
-to work well, there is no guarantee they are correct). What they are 
-best used for is to make a good guess regarding the data structure and then 
+to work well, there is no guarantee they are correct). What they are
+best used for is to make a good guess regarding the data structure and then
 let the user confirm it.
 
-As such there is a 'wizard' to aid in this process (use this in lieu of 
+As such there is a 'wizard' to aid in this process (use this in lieu of
 the above code - requires wxPython):
 
    from DSV import DSV
@@ -38,7 +38,7 @@ Requires Python 2.0 or later
 Wizards tested with wxPython 2.2.5/NT 4.0, 2.3.2/Win2000 and Linux/GTK (RedHat 7.x)
 """
 
-# VERSION 1.3.3
+VERSION = "1.3.7"
 
 """
 Bugs/Caveats:
@@ -66,30 +66,54 @@ To do:
      use any character as a text qualifier, not just quotes.
    - Finish wizards and move them into separate module.
    - Better guessHeaders() algorithm, although this is difficult.
-
+   - Optimize maps() - try to eliminate lambda when possible
+   - Optimize memory usage.  Presently the entire file is loaded and then saved as
+     a list.  A better approach might be to analyze a smaller part of the file and
+     then return an iterator to step through it.
 """
 
-# Changes from 1.3 to 1.3.1:
+# Changelog
+# 1.37
+#   - Fix for font issue under GTK2 (thanks to Ahmad Baitalmal)
+#   - Added some space below the Ok/Cancel buttons.
+#
+# 1.36
+#   - Bugfix submitted by "nobody" ;) on SF
+#
+# 1.3.4 to 1.3.5:
+#   - Nigel Hathaway finds yet another bug (or two).  Can't seem to make him
+#     use something else, so they had to be fixed.  It's especially difficult
+#     to ignore him since he provided the fix.  Very annoying.
+#     - Problem with odd quote/delimiter combinations (SF bug #620284)
+#     - Losing empty fields at beginning/end (#619771)
+#     - Whitespace stripped from around string (#620115)
+#
+# 1.3.3 to 1.3.4(a):
+#   - Fixed bug in exportDSV that failed to quote data containing delimiter
+#     thanks to nhathaway@users.sourceforge.net
+#
+# 1.3 to 1.3.1:
 #   - Test for presence of wxPython (since it's not required except for wizard)
 #   - Changed "from wxPython.wx import *" to "from wxPython import wx"
 #   - Changed sample csv file (darkwave.csv) to demonstrate embedded quotes
 
-
 import sys
 # import pre as re # sre was broken, appears okay now. Try this if there are problems.
 import re 
 import copy
 import exceptions
 
+# RedHat 8.0 (or rather GTK2?) sets LANG = en_us.UTF-8 and apparently some
+# older apps (including wxGTK) can't handle this.  The fix is to set LANG = C
+# before running the app.  Thanks to Ahmad Baitalmal for supplying this info.
+import os
+os.putenv('LANG', 'C')
+
 try:
-    from wxPython      import wx
-    from wxPython      import grid
+    from wxPython import wx, grid
 except ImportError:
     wx = None
 
-true = 1
-false = 0
-
 class InvalidDelimiter(exceptions.StandardError): pass
 class InvalidTextQualifier(exceptions.StandardError): pass
 class InvalidData(exceptions.StandardError): pass
@@ -357,7 +381,7 @@ def organizeIntoLines(input, textQualifier = '"', limit = None):
     # Algorithm: there should be an even number of text qualifiers on every line.
     # If there isn't, that means that the newline at the end of the line must occur
     # within qualifiers and doesn't really indicate the end of a record.
-
+    
     data = input.split('\n')
     line = 0
     while 1:
@@ -428,29 +452,69 @@ def importDSV(input, delimiter = ',', textQualifier = '"', columns = 0,
     if type(delimiter) != type('') or not delimiter:
         raise InvalidDelimiter, `delimiter`
 
-    if textQualifier:
-        fieldRex = re.compile('(?:(?:[%s]|^)%s(.*?)%s(?=[%s]|$))|(?:(?:[%s]|^)([^%s%s]*?)(?=[%s]|$))'
-                              % (delimiter, textQualifier, textQualifier, delimiter,
-                                 delimiter, textQualifier, delimiter, delimiter),
-                              re.S)
-    else:
-        fieldRex = re.compile('(?:[%s]|^)([^%s]*?)(?=[%s]|$)'
-                              % (delimiter, delimiter, delimiter), re.S)
+##    if textQualifier:
+##        # fieldRex=re.compile('(?:(?:[,]|^)"(.*?)"(?=[,]|$))|(?:(?:[,]|^)([^",]*?)(?=[,]|$))')
+##        fieldRex = re.compile('(?:(?:[%s]|^)%s(.*?)%s(?=[%s]|$))|(?:(?:[%s]|^)([^%s%s]*?)(?=[%s]|$))'
+##                              % (delimiter, textQualifier, textQualifier, delimiter,
+##                                 delimiter, textQualifier, delimiter, delimiter),
+##                              re.S)
+##    else:
+##        fieldRex = re.compile('(?:[%s]|^)([^%s]*?)(?=[%s]|$)'
+##                              % (delimiter, delimiter, delimiter), re.S)
 
     percent = 0.0
     lineno = 0.0
     newdata = []
     maxColumns = 0
 
-    for line in input:
-        line = line.strip()
-        record = fieldRex.findall(line)
-        if textQualifier:
-            record = [(i[0] or i[1]) for i in record]
+##    for line in input:
+##        line = line.strip()
+##        record = fieldRex.findall(line)
+##        print record
+##        if textQualifier:
+##            record = [(i[0] or i[1]) for i in record]
 
+##        if textQualifier:
+##            record = [c.replace(textQualifier * 2, textQualifier) for c in record]
+##        newdata.append(record)
+
+    # This code was submitted by Nigel to replace the code commented out above.
+    # It addresses several issues with embedded quotes and delimiters.  It seems that
+    # while off to a good start, regular expressions won't be able to handle certain
+    # situations. i.e. '''"Say ""hello"", World", ""''' would seem to be a problem as
+    # an embedded delimiter follows an embedded quote which throws off the re search.
+
+    for line in input:
         if textQualifier:
-            record = [c.replace(textQualifier * 2, textQualifier) for c in record]
+            record = []
+            inquotes = 0
+            for s in line.split(delimiter):
+                odd = s.count(textQualifier) % 2
+                if inquotes:
+                    accu += delimiter + s.replace(textQualifier * 2, delimiter).\
+                            replace(textQualifier, '').replace(delimiter, textQualifier)
+                    if odd:
+                        record.append(accu)
+                        inquotes = 0
+                else:
+                    # 1.3.6 bugfix: deal with case where s = "" to denote an empty string
+                    if s.count(textQualifier): # discard whitespace outside of textQualifiers when they are used
+                        s = s.strip()
+                    # fix new problem with ""
+                    if s == textQualifier * 2: 
+                        s = ""
+
+                    accu = s.replace(textQualifier * 2, delimiter).\
+                           replace(textQualifier, '').replace(delimiter, textQualifier)
+                    if odd:
+                        inquotes = 1
+                    else:
+                        record.append(accu)
+        else:
+            record = map(lambda x: x.strip(), line.split(delimiter))
+
         newdata.append(record)
+        # (end of replacement code)
         
         if updateFunction is not None:
             lineno = lineno + 1.0
@@ -508,10 +572,10 @@ def exportDSV(input, delimiter = ',', textQualifier = '"', quoteall = 0):
                    map(lambda j, q = q: q + j + q, i),
                    data)
     else: # quote only the values that contain qualifiers, delimiters or newlines
-        data = map(lambda i, q = textQualifier:
-                   map(lambda j, q = q: ((j.find(q) != -1 or j.find('\n') != -1)
-                                         and (q + j + q)) or j,
-                       i), data)
+        data = map(lambda i, q = textQualifier, d = delimiter:
+                   map(lambda j, q = q, d = d: ((j.find(q) != -1 or j.find(d) != -1
+                                          or j.find('\n') != -1)
+                                         and (q + j + q)) or j, i), data)
     # assemble each line with delimiters
     data = [delimiter.join(line) for line in data]
     
@@ -556,12 +620,12 @@ if wx is not None:
                      pos = wx.wxDefaultPosition, size = wx.wxDefaultSize,
                      style = wx.wxTAB_TRAVERSAL, name = "ImportWizardPanel"):
             wx.wxPanel.__init__(self, parent, id, pos, size, style, name)
-            self.SetAutoLayout(true)
+            self.SetAutoLayout(wx.true)
             mainSizer = wx.wxFlexGridSizer(3, 1)
             self.SetSizer(mainSizer)
             mainSizer.AddGrowableCol(0)
 
-            self.initialized = false
+            self.initialized = wx.false
             self.data = data
             self.isValidCallback = isValidCallback
             self.Validate = (isValidCallback and self.Validate) or self.BuildPreview
@@ -613,9 +677,9 @@ if wx is not None:
             wx.EVT_TEXT(self, self.otherDelim.GetId(), self.OnCustomDelim)
 
             if self.delimChecks.has_key(delimiter):
-                self.delimChecks[delimiter].SetValue(true)
+                self.delimChecks[delimiter].SetValue(wx.true)
             elif delimiter is not None:
-                self.delimChecks['Other'].SetValue(true)
+                self.delimChecks['Other'].SetValue(wx.true)
                 self.otherDelim.SetValue(delimiter)
 
             otherSizer.AddMany([
@@ -627,7 +691,7 @@ if wx is not None:
             delimStaticSizer.Add(delimGridSizer, 1, wx.wxEXPAND)
             delimOtherSizer = wx.wxBoxSizer(wx.wxVERTICAL)
             self.consecutiveDelimsAs1 = wx.wxCheckBox(self, -1, "Treat consecutive delimiters as one")
-            self.consecutiveDelimsAs1.Enable(false)
+            self.consecutiveDelimsAs1.Enable(wx.false)
             tqSizer = wx.wxBoxSizer(wx.wxHORIZONTAL)
             self.textQualifierChoice = wx.wxChoice(self, -1, choices = ['"', "'", "{None}"])
             wx.EVT_CHOICE(self, self.textQualifierChoice.GetId(), self.BuildPreview)
@@ -704,12 +768,12 @@ if wx is not None:
             previewStaticSizer = wx.wxStaticBoxSizer(previewStaticBox, wx.wxVERTICAL)
             self.preview = grid.wxGrid(self, -1)
             self.preview.CreateGrid(self.displayRows, cols)
-            self.preview.SetDefaultRowSize(self.preview.GetCharHeight() + 4, true)
-            self.preview.EnableEditing(false)
+            self.preview.SetDefaultRowSize(self.preview.GetCharHeight() + 4, wx.true)
+            self.preview.EnableEditing(wx.false)
             self.preview.SetColLabelSize(0)
             self.preview.SetRowLabelSize(0)
             self.preview.SetMargins(1, 0)
-            self.initialized = true
+            self.initialized = wx.true
             self.BuildPreview()
 
             rowheight = self.preview.GetRowSize(0) + 2
@@ -788,7 +852,7 @@ if wx is not None:
                 for col in range(cols):
                     try:    self.preview.SetColLabelValue(col, str(previewData[0][col]))
                     except: self.preview.SetColLabelValue(col, "")
-                # self.preview.AutoSizeColumns(true) # size columns to headers
+                # self.preview.AutoSizeColumns(wx.true) # size columns to headers
             else:
                 self.preview.SetColLabelSize(0)
 
@@ -798,7 +862,7 @@ if wx is not None:
                     except: pass
 
             # if not hasHeaders:
-            self.preview.AutoSizeColumns(true) # size columns to data
+            self.preview.AutoSizeColumns(wx.true) # size columns to data
 
             rowheight = self.preview.GetRowSize(0)
             self.preview.SetRowSize(0, rowheight)
@@ -854,7 +918,7 @@ if wx is not None:
                      pos = wx.wxDefaultPosition, size = wx.wxDefaultSize,
                      style = wx.wxDEFAULT_DIALOG_STYLE, name = "ImportWizardDialog"):
             wx.wxDialog.__init__(self, parent, id, title, pos, size, style, name)
-            self.SetAutoLayout(true)
+            self.SetAutoLayout(wx.true)
 
             self.file = file
             f = open(file, 'r')
@@ -877,14 +941,14 @@ if wx is not None:
 
         def ButtonBox(self):
             panel = wx.wxPanel(self, -1)
-            panel.SetAutoLayout(true)
+            panel.SetAutoLayout(wx.true)
             sizer = wx.wxBoxSizer(wx.wxHORIZONTAL)
             panel.SetSizer(sizer)
             self.ok = wx.wxButton(panel, wx.wxID_OK, "Ok")
             cancel = wx.wxButton(panel, wx.wxID_CANCEL, "Cancel")
             sizer.AddMany([
-                (self.ok, 0, wx.wxALIGN_TOP | wx.wxEAST, 5),
-                (cancel, 0, wx.wxALIGN_TOP | wx.wxWEST, 5),
+                (self.ok, 0, wx.wxALIGN_TOP | wx.wxEAST | wx.wxSOUTH, 10),
+                (cancel, 0, wx.wxALIGN_TOP | wx.wxWEST | wx.wxSOUTH, 10),
                 ])
             panel.Layout()
             panel.Fit()
@@ -942,7 +1006,7 @@ if __name__ == '__main__':
             # log the bad row to a file
             file.write(oldrow + '\n')
         
-        dlg = ImportWizardDialog(None, -1, 'CSV Import Wizard', path)
+        dlg = ImportWizardDialog(None, -1, 'CSV Import Wizard (v.%s)' % VERSION, path)
         if dlg.ShowModal() == wx.wxID_OK:
             results = dlg.ImportData(errorHandler = logErrors)
             dlg.Destroy()
@@ -955,6 +1019,11 @@ if __name__ == '__main__':
                     print 80*'='
                 for row in data:
                     print row
-                    
+
+                if headers:
+                    print exportDSV([headers] + data)
+                else:
+                    print exportDSV(data)
+            
     else:
         dlg.Destroy()
